const fs = require('fs')
const md5 = require('md5')
const { promisify } = require('util')
const PythonWorker = require('./pythonWorker')

const readdir = promisify(fs.readdir)

const state = {
  workers: [],
  interfaces: [],
  ports: [],
  connections: []
}

const typeDefs = `
  type Option {
    name: String!
    type: String!
    description: String
    values: [String!]
  }

  type Port {
    id: String!
    device: String!
    interfaceName: String!
    name: String
    description: String
  }

  type Connection {
    id: ID!
    interfaceName: String!
    device: String!
  }

  type Worker {
    interfaceName: String!
    workerScript: String!
  }

  type Interface {
    interfaceName: String!
    workerScript: String!
    ports: [Port]!
    connections: [Connection]!
    options: [Option]!
  }

  extend type Query {
    interfaces(force: Boolean): [Interface]!
    ports(interfaceName: String!, force: Boolean): [Port]!
    connections: [Connection]!
  }

  extend type Mutation {
    connect(interfaceName: String!, device: String!): Connection!
    connectionCommand(connectionId: ID!, type: String!, string: String!, options: String): String!
  }
`

async function findWorkers () {
  // Find all files in ./python_workers that end in _worker.py
  const fileNames = await readdir(`${__dirname}/python_workers`)
  const workerFiles = fileNames.filter(fileName => fileName.includes('_worker.py'))

  // Find the added workers
  workerFiles.forEach(workerFile => {
    const interfaceName = workerFile.replace(/_worker\.py/, '')
    if (state.workers.find(worker => worker.interfaceName === interfaceName)) return null
    const workerScript = `${__dirname}/python_workers/${workerFile}`
    state.workers.push({
      interfaceName,
      workerScript
    })
  })
}

async function findInterfaces () {
  // Try to identify workers if necessary
  if (!state.workers.length) await findWorkers()

  // Generate an interface for each worker
  state.workers.forEach(worker => {
    const { workerScript, interfaceName } = worker
    // Skip existing interfaces
    if (state.interfaces.find(iface => iface.interfaceName === interfaceName)) return null
    state.interfaces.push({
      interfaceName,
      workerScript,
      worker: new PythonWorker(workerScript),
      ports: [],
      connections: [],
      options: []
    })
  })
}

async function interfaces (parent, args, ctx, info) {
  const { force } = args
  // Try to identify interfaces if necessary
  if (!state.interfaces.length || force) await findInterfaces()

  return state.interfaces.map(iface => {
    const { workerScript, interfaceName } = iface
    return {
      interfaceName,
      workerScript,
      ports: ports(interfaceName, args, ctx, info),
      connections: connections(interfaceName, args, ctx, info),
      options: options(interfaceName, args, ctx, info)
    }
  })
}

async function findPorts (interfaceName) {
  // Generate all ports for the interface
  const iface = state.interfaces.find(iface => iface.interfaceName === interfaceName)

  const { data, error } = await iface.worker.send({ type: 'ports' })
  if (error) throw new Error(JSON.stringify(error))
  data.forEach(port => {
    const id = port.name || port.device
    // Skip existing ports
    if (state.ports.find(port => port.id === id)) return null
    const newPort = {
      id,
      interfaceName,
      ...port
    }
    state.ports.push(newPort)
    iface.ports.push(newPort)
  })
}

async function ports (parent, args, ctx, info) {
  const { force, interfaceName } = args
  const ifName = interfaceName || parent

  if (ifName) {
    const iface = state.interfaces.find(iface => iface.interfaceName === ifName)

    // Try to find ports if necessary
    if (!iface.ports.length || force) await findPorts(ifName)
    return iface.ports
  } else {
    return state.ports
  }
}

async function findOptions (interfaceName) {
  const iface = state.interfaces.find(iface => iface.interfaceName === interfaceName)
  const { data, error } = await iface.worker.send({ type: 'options' })
  if (error) throw new Error(JSON.stringify(error))
  iface.options.push(...data)
}

async function options (parent, args, ctx, info) {
  const iface = state.interfaces.find(iface => iface.interfaceName === parent)

  // Try to find options if necessary
  if (!iface.options.length) await findOptions(parent)
  return iface.options
}

async function connections (parent, args, ctx, info) {
  if (parent) {
    const iface = state.interfaces.find(iface => iface.interfaceName === parent)
    return iface.connections
  } else {
    return state.connections
  }
}

async function connect (parent, args, ctx, info) {
  const { interfaceName, device } = args
  const iface = state.interfaces.find(iface => iface.interfaceName === interfaceName)
  const id = md5(interfaceName + device)
  if (iface.connections.find(connection => connection.id === id)) throw new Error('already connected.')
  const connection = {
    id,
    device,
    interfaceName,
    worker: new PythonWorker(iface.workerScript)
  }
  const { error } = await connection.worker.send({ type: 'connect', device })
  if (error) throw new Error(error)
  iface.connections.push(connection)
  state.connections.push(connection)
  return connection
}

async function connectionCommand (parent, args, ctx, info) {
  const { connectionId, type, string, options } = args
  const connection = state.connections.find(connection => connection.id === connectionId)
  const { data } = await connection.worker.send({ type, string, options })
  return data.response
}

const resolvers = {
  Query: {
    interfaces,
    ports,
    connections
  },
  Mutation: {
    connect,
    connectionCommand
  }
}

module.exports = { typeDefs, resolvers }
